#!/usr/bin/env ./localpython/bin/python

import os
import sys
import numpy
import tensor
import sptensor
import tenmat
import scipy
import scipy.stats
import scipy.special
import scipy.optimize
import beta_ntf
import tensor_fac
import time

''' kernel parameter '''
class KernelParam:
    def __init__(self,name):
        self.name = name
        self.kreg = 1e-6
        self.tol = 1e-3
        self.nu = 0.5 #matern cov
        self.phi = 1.0 #matern cov
        self.alpha = 1.0 #matern cov
        self.gamma = 1.0 #rbf cov
        self.degree = 2 #poly cov
        self.taper = 'spherical' #taper function
        self.theta = 1.0 #taper function


''' object to store gram matrix and their eigen values/vectors (bigger ones)'''
class Kernel:
    def __init__(self):
        self.eig_vals = None
        self.eig_vecs = None
        self.K = None

''' Object to store kernel function for tensor'''
class TensorKernel:
    def __init__(self, nmod):
        self.nmod = nmod
        self.K = [None for i in range(nmod)]


class Model:
    def __init__(self):
        self.var_total = 50
        self.var_em_total = 15
        self.stream_total = 5000
        '''learning rate related '''
        self.kai = 1
        self.tao0 = 4096
        self.eta0 = 0.1
        self.subsample_siz = None
        self.dim = 3
        self.offset = 2.0
        ''' learning strategy
            0: update all modes;
            1: update each mode cyclicly
            2: update all modes, with L1
            3: update each mode cyclicly, with L1
        '''
        self.strategy = 1
        self.lam = 0.1
        self.U = None

class Data:
    def __init__(self):
        self.nmod = None
        self.nvec = None
        self.Y = None
        self.nzind = None
        self.test_ind = None
        self.test_y = None


''' Caclulate the gram matrix of the n*p input matrix, where each row is a data point'''
def kernel_func(X, param):
    n,p = X.shape
    K = None
    if param.name == 'linear':
        K = numpy.dot(X,X.transpose())
    elif param.name == 'rbf':
        norm2 = numpy.sum(X**2,axis=1)
        norm2 = numpy.tile(norm2, [n,1])
        K = norm2 + norm2.transpose() - 2*numpy.dot(X, X.transpose())
        K = numpy.exp(-param.gamma*K)
    elif param.name == 'poly':
        K = numpy.dot(X,X.transpose())**param.degree
    elif param.name == 'taper_matern':
        if param.taper == 'spherical' and param.nu>0.5:
            print 'For spherical taper function, nu must <=0.5'
            return None
        if param.taper == 'wendland1' and param.nu>1.5:
            print 'For Wendland1 taper function, nu must <=1.5'
            return None
        if param.taper == 'wendland2' and param.nu>2.5:
            print 'For Wendland2 taper function, nu must <=2.5'
            return None
        norm2 = numpy.sum(X**2,axis=1)
        norm2 = numpy.tile(norm2, [n,1])
        norm2 = norm2 + norm2.transpose() - 2*numpy.dot(X, X.transpose())
        norm2[norm2<0] = 0.0
        H = numpy.sqrt(norm2).real.reshape(norm2.size)

        if param.taper == 'spherical':
            covTap = ((H < param.theta)*(1 - H/param.theta))**2*(1 + H/(2*param.theta))
        elif param.taper == 'wendland1':
            covTap = ((H < param.theta)*(1 - H/param.theta))**4*(1 + 4*H/param.theta)
        elif param.taper == 'wendland2':
            covTap = ((H < param.theta)*(1 - H/param.theta))**6*(1 + 6*H/param.theta + 35*H**2/(3*param.theta**2))
        else:
            print 'Unkown tapering function'
            return None
        nzH = H[covTap>1e-8] + 1e-8
        nzCovMatern = param.phi/(2**(param.nu-1)*scipy.special.gamma(param.nu))*(param.alpha*nzH)**param.nu*scipy.special.kv(param.nu, param.alpha*nzH)
        nzCovMatern[numpy.isnan(nzCovMatern)] = 1.0
        K = numpy.zeros(n*n)
        K[covTap>1e-8] = nzCovMatern*covTap[covTap>1e-8]
        K = K.reshape([n,n])
    else:
        print 'Unkown kernel types'
        return None

    K = K + param.kreg*numpy.eye(n)

    #here we compress the eigen-decomposition of K by throwing out the smallest eigen values and eigen vectors
    [eig_vals, eig_vecs] = numpy.linalg.eig(K)
    eig_val_threshold  = numpy.max(eig_vals)*param.tol
    ret = Kernel()
    ret.K = K
    ret.eig_vals = eig_vals[eig_vals>eig_val_threshold]
    ret.eig_vals = ret.eig_vals.reshape([ret.eig_vals.size, 1])
    ret.eig_vecs = eig_vecs[:,eig_vals>eig_val_threshold]
    return ret

'''X must be a list of arrays, each arry is a N_k*p_k latent representation of the nodes in k-th mode '''
def tensor_kernel_func(X, param):
    ret = TensorKernel(len(X))
    for i in range(ret.nmod):
        ret.K[i] = kernel_func(X[i],param)
    return ret

def taper_matern_grad_U(U, param):
    n,dim = U.shape
    norm2 = numpy.sum(U**2,axis=1)
    norm2 = numpy.tile(norm2, [n,1])
    norm2 = norm2 + norm2.transpose() - 2*numpy.dot(U, U.transpose())
    norm2[norm2<0] = 0.0
    H = numpy.sqrt(norm2).real.reshape(norm2.size)

    tapered = (H < param.theta)*(1 - H/param.theta)
    if param.taper == 'spherical':
        covTap = tapered**2*(1 + H/(2*param.theta))
        covTapGrad = -2.0/param.theta*tapered*(1+H/(2*param.theta)) + tapered**2/(2*param.theta)
    elif param.taper == 'wendland1':
        covTap = tapered**4*(1.0 + 4*H/param.theta)
        covTapGrad = -4.0/param.theta*tapered**3*(1+4*H/param.theta) + tapered**4*4/param.theta
    elif param.taper == 'wendland2':
        covTap = tapered**6*(1 + 6*H/param.theta + 35*H**2/(3*param.theta**2))
        covTapGrad = -6.0/param.theta*tapered**5*(1.0+6*H/param.theta+35*H**2/3/param.theta**2) + tapered**6*(6.0/param.theta+70*H/3/param.theta**2)
    else:
        print 'Unkown tapering function'
        return None

    #nzH = H[covTap>1e-8] + 1e-8
    nzH = H[H<param.theta] + 1e-8
    nzH = nzH*(numpy.abs(nzH)>1e-8)
    nzCovMatern = param.phi/(2**(param.nu-1)*scipy.special.gamma(param.nu))*(param.alpha*nzH)**param.nu*scipy.special.kv(param.nu, param.alpha*nzH)
    nzCovMatern[numpy.isnan(nzCovMatern)] = 1.0
    nzCovMatern[numpy.isinf(nzCovMatern)] = 1.0


    bkgrad = -0.5*(scipy.special.kv(param.nu-1,param.alpha*nzH) + scipy.special.kv(param.nu+1,param.alpha*nzH))
    bkgrad[numpy.isnan(bkgrad)] = 1.0
    bkgrad[numpy.isinf(bkgrad)] = 1.0

    nzCovMaternGrad = param.phi/(2**(param.nu-1)*scipy.special.gamma(param.nu)) \
            * (param.nu*param.alpha*(param.alpha*nzH)**(param.nu-1)*scipy.special.kv(param.nu,param.alpha*nzH) \
            + param.alpha*(param.alpha*nzH)**param.nu*bkgrad)
    nzCovMaternGrad[numpy.isnan(nzCovMaternGrad)] = 0.0
    nzCovMaternGrad[numpy.isinf(nzCovMaternGrad)] = 0.0

    gKh = numpy.zeros(n*n)
    gKh[H<param.theta] = nzCovMaternGrad*covTap[H<param.theta] + nzCovMatern*covTapGrad[H<param.theta]
    gKh = gKh.reshape([n,n])

    gdU = [None for i in range(dim)]
    for i in range(dim):
        repU = numpy.tile(U[:,i].reshape(n,1),[1,n])
        coefs =  (repU - repU.transpose()).reshape(n*n)
        #coefs[H>1e-8] = coefs[H>1e-8]/H[H>1e-8]
        #coefs[H<=1e-8] = 0.0
        coefs = coefs/(H+1e-8)
        coefs = coefs.reshape([n,n])
        gdU[i] = coefs*gKh
    return gdU

''' Ukvec: a long vector (array) represent \U_k
    k: k-th mode
    Utmsnok: a tensor
    eigval_list: a list, where each element is a array of eigen values of gram-matrix of each mode
    eigvec_k: eigen vectors of k-th mode
    pram: kernel parameters
    Uk0: the prior
    sigma: the variance parameter
'''
def tensor_gradient_U(Ukvec, k, Utmsnok,  eigval_list, eigvec_k, param, Uk0, sigma, num_points):
    nvec = Utmsnok.shape
    n = Utmsnok.size()
    nmod = len(nvec)
    nk = nvec[k]

    U_k = Ukvec.reshape([nk, Ukvec.size/nk])
    kernel_k = kernel_func(U_k, param)
    inv_k_half = numpy.dot(numpy.dot(kernel_k.eig_vecs, numpy.diagflat(kernel_k.eig_vals**(-0.5))), kernel_k.eig_vecs.transpose())
    sqroot_term = Utmsnok.ttm(inv_k_half,k)

    core = tensor.tensor([1],[1 for i in range(nmod)])
    DD = core.ttm(eigval_list, [i for i in range(nmod)])
    DD = tensor.tensor(DD.data/(1.0+DD.data), DD.shape)
    inv_eigvals = [1.0/x.transpose() for x in eigval_list]
    KV = numpy.linalg.solve(kernel_k.K, eigvec_k)
    inv_eigvals[k] = numpy.dot(eigvec_k.transpose(),KV).diagonal().reshape(eigval_list[k].shape[::-1])

    fun_val = 0.5*n/nk*numpy.sum(numpy.log(kernel_k.eig_vals))  + 0.5*numpy.sum(sqroot_term.data**2) + \
            0.5*numpy.sum(DD.ttm(inv_eigvals,[i for i in range(nmod)]).data) + \
            0.5/num_points/sigma*numpy.sum((U_k - Uk0)**2)

    g = numpy.zeros(U_k.shape)
    d = U_k.shape[1]
    A = tenmat.tenmat(Utmsnok,[k]).data
    KM = numpy.linalg.solve(kernel_k.K, numpy.dot(A,A.transpose()))
    modnok = numpy.array(range(nmod))
    modnok = modnok[modnok!=k]
    if param.name == 'taper_matern':
        gradU = taper_matern_grad_U(U_k, param)
    for j in range(0,d):
        repU = numpy.tile(U_k[:,j].reshape([nk,1]),[1,nk])
        if param.name == 'rbf':
            Kd = -2*param.gamma*numpy.linalg.solve(kernel_k.K, kernel_k.K*(repU.transpose() - repU))
        elif param.name == 'linear':
            Kd = numpy.linalg.solve(kernel_k.K, repU)
        elif param.name == 'poly':
            Kd = param.degree*numpy.linalg(kernel_k.K, kernel_k.K/numpy.dot(U_k, U_k.transpose())*repU)
        elif param.name == 'taper_matern':
            Kd = numpy.linalg.solve(kernel_k.K, gradU[j])
        else:
            print 'Unkown kernel types'
            return None
        b = DD.ttm(inv_eigvals, modnok).data.reshape([eigval_list[k].size,1])
        gCol = n/nk*Kd.diagonal().reshape(nk,1) - numpy.sum(KM.transpose()*Kd, axis=0).reshape(nk,1) - \
               numpy.dot(numpy.dot(eigvec_k.transpose(),Kd).transpose() * KV, b)
        g[:,j] = gCol.reshape(nk).real
    g = g + (U_k-Uk0)/(num_points*sigma)
    #g = g + param.kreg  #forget about the prior
    #g = g/numpy.sqrt(numpy.sum(g**2))
    return [fun_val.real,g.real]


''' E step for infiniteTucker
    M: n-d array, the latent value for each tensor entry
    Y: n-d array, the observed tensor entry value
    model: model setting atributes
    model.offset: the offset for probit model
    tensor_kern: the calculated TensorKernel object
'''
def var_E(M, Y, tensor_kern, model):
    tol = 1e-10
    nvec = Y.shape
    nmod = len(nvec)
    nmod_arr = [i for i in range(nmod)]
    nmod_one = [1 for i in range(nmod)]
    eigval_list = [x.eig_vals for x in tensor_kern.K]
    eigvec_list = [x.eig_vecs for x in tensor_kern.K]
    eigvecT_list = [x.eig_vecs.transpose() for x in tensor_kern.K]
    core = tensor.tensor([1],nmod_one)
    DD = core.ttm(eigval_list, nmod_arr)
    DD = tensor.tensor(DD.data/(1.0+DD.data), DD.shape)
    if model.loss == 'Gauss':
        M = tensor.tensor(Y.ttm(eigvecT_list, nmod_arr).tondarray()*DD.data).ttm(eigvec_list, nmod_arr).data.real
    else:
        for iter_num in range(model.var_total):
            P = (2*Y - 1)*scipy.stats.norm.pdf(M-model.offset)/(scipy.stats.norm.cdf((2*Y-1)*(M-model.offset)) + tol)
            Z = tensor.tensor(P + M, P.shape)
            A = Z.ttm(eigvecT_list, nmod_arr).data*DD.data
            A = tensor.tensor(A,A.shape)
            M = A.ttm(eigvec_list, nmod_arr).data.real
        #print iter_num

    return M



def test_online():
    data_arr = numpy.loadtxt('../onlineInfTucker/data.txt').astype(int)
    #test_sub = numpy.loadtxt('../onlineInfTucker/test_sub.txt').astype(int)
    #test_y = numpy.loadtxt('../onlineInfTucker/test_y.txt').astype(int)
    test_data = numpy.loadtxt('../onlineInfTucker/test.txt')
    test_sub = test_data[:,0:3] - 1
    test_y = test_data[:,-1]

    data = Data()
    data.nvec = data_arr[0,:]
    data.nmod = len(data.nvec)
    data.nzind = data_arr[1:,:] - 1
    data.Y = sptensor.sptensor(data.nzind,numpy.ones([data.nzind.shape[0],1]),data.nvec)
    data.test_ind = test_sub
    data.test_y = test_y


    model = Model()
    model.dim = 5
    model.subsample_siz = numpy.array([40,40,40])
    model.eta0 = 0.1
    model.tao0 = 2048
    model.strategy = 3

    ker_param = KernelParam('taper_matern')

    U = online_learning(data, model, ker_param)

'''
data.subinds_arr contains the sub-tensor set
'''
def sgd_learning(data, model, ker_param):

    U = model.U
    '''
    model.sgd_iter_total: total number for stochastic gradient descent
    model.subinds_arr: list of sub-tensors
    model.learning_rate: const learning rate
    '''
    for iter_num in xrange(model.sgd_iter_total):
        #randomly shuffle the data points
        perm = numpy.random.permutation(len(data.subinds_arr))
        perm = perm[0:model.sgd_sample_number]
        count = 0
        #__sttime = time.time()
        for t in perm:
            count += 1
            #print t
            #__sttime = time.time()
            subinds = data.subinds_arr[t]
            model.subsample_siz = [len(x) for x in subinds]
            U_sub = [U[i][subinds[i],:]   for i in range(data.nmod) ]
            U_sub0 = [model.U0[i][subinds[i],:]   for i in range(data.nmod) ]
            #print subinds[0].size,subinds[1].size,subinds[2].size
            nz_ind =  data.nzind
            if model.loss == 'Gauss':
                nz_vals = data.nzvals
            for k in range(data.nmod):
                bool_indicator = numpy.in1d(nz_ind[:,k], subinds[k])
                nz_ind = nz_ind[bool_indicator,:]
                if model.loss == 'Gauss':
                    nz_vals = nz_vals[bool_indicator]
                if nz_ind.size == 0:
                    break
            #if nz_ind.size == 0:
            #    print 'empty'
            #get the reverse map
            rmap = [None for i in range(data.nmod)]
            for k in range(data.nmod):
                rmap[k] = numpy.zeros(numpy.max(subinds[k]) + 1)
                rmap[k][subinds[k]] = numpy.arange(subinds[k].size)
            ind_sub = numpy.zeros(nz_ind.shape)
            for k in range(data.nmod):
                ind_sub[:,k] = rmap[k][nz_ind[:,k]]

            if model.loss == 'Gauss':
                Y_sub_tensor = sptensor.sptensor(ind_sub, nz_vals.reshape([ind_sub.shape[0],1]), model.subsample_siz)
            else:
                Y_sub_tensor = sptensor.sptensor(ind_sub, numpy.ones([ind_sub.shape[0],1]), model.subsample_siz)
            Y_sub = Y_sub_tensor.totensor().data
            if nz_ind.size == 0:
                M_sub = numpy.zeros(model.subsample_siz)
            else:
                if model.loss == 'Gauss':
                    M_sub = Y_sub
                else:
                    '''
                    ntf = beta_ntf.BetaNTF(Y_sub.shape,n_components=model.dim[0],beta = 2,n_iter = 10,verbose=False);
                    ntf.fit(Y_sub)
                    M_sub = beta_ntf.parafac(ntf.factors_)
                    '''
                    tfac = tensor_fac.TensorFac()
                    [core, U_sub_tmp] = tfac.tucker_als(Y_sub_tensor, model.dim)
                    M_sub = core.ttm(U_sub_tmp, numpy.arange(len(U_sub_tmp))).data
                    #M_sub = 2*Y_sub-1

                    if nz_ind.size > 0:
                        M_sub = M_sub/numpy.sqrt(numpy.sum(M_sub**2))*numpy.sqrt(numpy.sum(Y_sub))
            #E step
            tensor_kern = tensor_kernel_func(U_sub, ker_param)
            if model.loss == 'Gauss':
                M_sub = var_E(M_sub, Y_sub_tensor.totensor(), tensor_kern, model)
            else:
                M_sub = var_E(M_sub, Y_sub, tensor_kern, model)

            #gradient
            eigvals = [x.eig_vals for x in tensor_kern.K]
            inv_half_ker = [numpy.dot(numpy.dot(x.eig_vecs, numpy.diagflat(x.eig_vals**(-0.5))),x.eig_vecs.transpose()) for x in tensor_kern.K]
            coef = model.learning_rate
            ''' update all the modes '''
            if model.strategy == 0:
                g = [None for i in range(data.nmod)]
                for k in range(data.nmod):
                    modnok = numpy.array(range(data.nmod))
                    modnok = modnok[modnok!=k]
                    Utmsnok = tensor.tensor(M_sub, M_sub.shape).ttm(inv_half_ker,modnok)
                    fun,g[k] = tensor_gradient_U(U_sub[k].reshape([U_sub[k].size,1]), k, Utmsnok, eigvals, tensor_kern.K[k].eig_vecs, ker_param, U_sub0[k], model.sigma, len(data.subinds_arr) )

                for k in range(data.nmod):
                    U_sub[k] = U_sub[k] - coef*g[k]
                    U[k][subinds[k],:] = U_sub[k]
            elif model.strategy == 1:
                '''update single mode cyclicly'''
                k = count%data.nmod
                modnok = numpy.array(range(data.nmod))
                modnok = modnok[modnok!=k]
                Utmsnok = tensor.tensor(M_sub, M_sub.shape).ttm(inv_half_ker,modnok)
                fun,g = tensor_gradient_U(U_sub[k].reshape([U_sub[k].size,1]), k, Utmsnok, eigvals, tensor_kern.K[k].eig_vecs, ker_param, U_sub0[k], model.sigma, len(data.subinds_arr))
                U_sub[k] = U_sub[k] - coef*g
                #print 't=%d, nan=%d'%(t, numpy.sum(numpy.isnan(U_sub[k])))
                #if numpy.sum(numpy.isnan(U_sub[k]))>0:
                #    print 'cao'
                U[k][subinds[k],:] = U_sub[k]

            else:
                print 'Unkown strategies'
                return None;

            #print 'cost %g'%(time.time() - __sttime)
        #print 'cost %g'%(time.time() - __sttime)
    return U


def mapper_job(data_arr, sub_inds, init_U, loss, learning_rate, iter_num):
    data = Data()
    if loss =='Gauss':
        ind_info = data_arr[0].astype(int)
        data.nvec = ind_info[0,:]
        data.nmod = len(data.nvec)
        data.nzind = ind_info[1:,:]
        data.nzvals = data_arr[1]
    else:
        data.nvec = data_arr[0,:].astype(int)
        data.nmod = len(data.nvec)
        data.nzind = data_arr[1:,:].astype(int)

    data.subinds_arr = []
    model = Model()
    model.dim = [x.shape[1] for x in init_U]
    model.U = [numpy.asarray(x) for x in init_U]
    model.U0 =  [x.copy() for x in model.U] #note: must be copy here
    model.sigma = 1
    model.sgd_iter_total = 1
    model.learning_rate = learning_rate
    model.strategy = 0
    model.loss = loss
    #only for binary
    model.var_total = 3

    subtensor_number = sub_inds[0]
    st = 1
    for i in xrange(subtensor_number):
        siz = []
        subinds = []
        for j in xrange(data.nmod):
            siz.append(sub_inds[st])
            inds = [sub_inds[i] for i in range(st + 1, st + sub_inds[st] + 1)]
            subinds.append(numpy.array(inds))
            st = st + sub_inds[st] + 1
        data.subinds_arr.append( subinds )
    #ker_param = KernelParam('taper_matern')
    model.sgd_sample_number = len(data.subinds_arr)/iter_num + 1
    ker_param = KernelParam('rbf')
    U = sgd_learning(data, model, ker_param)
    #output U
    for i in xrange(data.nmod):
        for j in xrange(data.nvec[i]):
            key = str(i) + '_' + str(j)
            val = ','.join([str(x) for x in U[i][j,:]])
            print key + ' ' + val


'''
    nvec: the dimensions of each mode
    dim: dimension of latent representation in each mode
    U_hdfs_path: U hdfs path
'''
def reduce_job(nvec, dim, U_hdfs_path):
    nmod = len(nvec)
    U = numpy.ndarray(nmod, dtype=numpy.dtype('object'))
    count = numpy.ndarray(nmod, dtype=numpy.dtype('object'))
    for i in range(nmod):
        U[i] = numpy.zeros([nvec[i],dim[i]])
        count[i] = numpy.zeros([nvec[i],dim[i]])

    #ff = open('tmp.red')
    #for line in ff:
    for line in sys.stdin:
        key,value = line.strip().split(' ')
        mode,nid = key.split('_')
        mode = int(mode)
        nid = int(nid)
        value = numpy.array([float(x) for x in value.split(',')])
        U[mode][nid,:] += value
        count[mode][nid,:] += 1.0

    #for map estimation, simply average is correct
    for mode in range(nmod):
        U[mode][count[mode][:,0]>0.9,:] /= count[mode][count[mode][:,0]>0.9,:]

    put_hdfs_data(U_hdfs_path, U)


#obj should be numpy ndarray
def put_hdfs_data(hdfs_path, obj):
    os.system('hadoop fs -rm ' + hdfs_path)
    tmp_name = hdfs_path.split('/')[-1]
    os.system('rm -f ' + tmp_name)
    numpy.save(tmp_name, obj)
    os.system('hadoop fs -put ' + tmp_name + ' ' + hdfs_path)
    os.system('rm -f ' + tmp_name)


def load_hdfs_data(hdfs_path):
    tmp_name  = hdfs_path.split('/')[-1]
    os.system('rm -f ' + tmp_name)
    os.system('hadoop fs -get ' + hdfs_path + ' ' + tmp_name)
    ret = numpy.load(tmp_name)
    os.system('rm -f ' + tmp_name)
    return ret

if __name__ == '__main__':
    if len(sys.argv) != 5 and len(sys.argv) != 7:
        print 'Usage:'
        print sys.argv[0] + ' -map <data_path>  <U_path> <loss> <learning rate> <iter_num> (e.g. -map /tmp/zsd/data.npy /tmp/zsd/U.npy Gauss 1e-8 5)'
        print sys.argv[0] + ' -red <nvec> <dim> <U_path> (e.g. -red 581,124,48 5,5,5 /tmp/zsd/U.npy)'
        sys.exit(0)

    old_settings = numpy.seterr(all='ignore')
    if sys.argv[1] == '-map':
        data_arr = load_hdfs_data(sys.argv[2])
        init_U = load_hdfs_data(sys.argv[3])
        loss = sys.argv[4]
        learning_rate = float(sys.argv[5])
        iter_num = int(sys.argv[6])

        #f = open('./input_small.txt')
        #for line in f:
        for line in sys.stdin:
            items = line.strip().split(' ')
            subinds = [int(x) for x in items]
            mapper_job(data_arr, subinds, init_U, loss, learning_rate, iter_num)
    else:
        ''' reduce job '''
        nvec = numpy.array([int(x) for x in sys.argv[2].split(',')])
        dim = numpy.array([int(x) for x in sys.argv[3].split(',')])
        U_hdfs_path = sys.argv[4]

        reduce_job(nvec, dim, U_hdfs_path)
